#!/bin/sh

#Exit immediately on error
set -e

# User multi core
CORES=`grep -c '^processor' /proc/cpuinfo`
CORES=$((CORES+1))
export J="-j${CORES}"

case $1 in
    clean | distclean)
		make clean
		rm -f crypto/aes/aes-armv4.S crypto/bn/armv4-gf2m.S
		rm -f crypto/modes/ghash-armv4.S crypto/sha/sha1-armv4-large.S
		rm -f crypto/sha/sha256-armv4.S crypto/sha/sha512-armv4.S
		rm -f crypto/aes/aesv8-armx.S crypto/aes/bsaes-armv7.S
		rm -f crypto/bn/armv4-mont.S crypto/modes/ghashv8-armx.S
		echo "Build products were cleaned."
		;;
    build | all)
		DIR=$2
		if [ -z "$DIR" ]; then
			if [ -z "$NDK_ROOT" ]; then
				NDK_ROOT="/opt/android/ndk-r10d"
			fi
		else
			NDK_ROOT=$DIR
		fi
		NDK_VERSION="`cat $NDK_ROOT/RELEASE.TXT`"
		if [ "$NDK_VERSION" = "r12b" ]; then
			COMPILER_VERSION="4.9"
			HOST=linux-x86_64
		else
			COMPILER_VERSION="4.8"
			HOST=linux-x86
		fi
		PLATFORM=android-16
		TOOLCHAIN=$NDK_ROOT/toolchains/arm-linux-androideabi-$COMPILER_VERSION/prebuilt/$HOST/bin/arm-linux-androideabi-
		echo "NDK_ROOT is set to $NDK_ROOT"
		echo "COMPILER_VERSION=$COMPILER_VERSION"
		echo "HOST=$HOST"
		echo "TOOLCHAIN=$TOOLCHAIN"

		CC=${TOOLCHAIN}gcc
		AR=${TOOLCHAIN}ar
		RANLIB=${TOOLCHAIN}ranlib
		ANDROID_DEV=$NDK_ROOT/platforms/$PLATFORM/arch-arm/usr
		export CC AR RANLIB ANDROID_DEV

		CIPHERS="no-md2 no-rc5 no-sha0 no-smime \
				 no-rmd160 no-aes192 no-ans1 no-krb5 \
				 no-capieng no-gmp no-jpake \
				 no-store no-rfc3779"
		OPTIONS="no-hw no-engines no-err no-fips no-sse2 no-perlasm zlib $EXTRA_OPTIONS"

		./Configure shared android-armv7 $CIPHERS $OPTIONS
		# other possible targets : android-armv7 linux-generic32

		make build_libs

		echo "Build products were builded."
		;;
    *)
		echo ""
		echo "usage: `basename $0` [ARGS]"
		echo ""
		echo "build|all [<ndk_directory>] : build using optional specified android NDK directory"
		echo "clean|distclean             : remove build products"
		echo ""
		exit 2
		;;
esac

